#!/usr/local/bin/python
# -*- coding: utf-8 -*-
# Python Closed Caption decoding - presented by notonbluray.com
# Code is released into the public domain
# But always appreciate attribution where possible
#
# All Trademarks are the property of their respective owners.
# See spec: http://www.gpo.gov/fdsys/pkg/CFR-2007-title47-vol1/pdf/CFR-2007-title47-vol1-sec15-119.pdf
#
# This code has no warranted, and used entirely at your own risk
#
import cv
import os
import subprocess
import tempfile
import time

# TODO - Make sure you set this !
PATH_TO_FFMPEG = 'X:\\Video\\tools\\ffmpeg.exe'

## Assumes 27 pixel wide 'bit' starting at pixel 280 - assumes 720 pixel wide video
## Odd parity on the rightmost bit, we sample central pixels of the bit and average
BYTE1_LOCATIONS = [ 285 + ( i * 27 ) for i in range( 0, 8 ) ] 
BYTE2_LOCATIONS = [ 285 + ( i * 27 ) for i in range( 8, 16 ) ]

## Sine wave preamble indicates presence of captions
SYNC_SIGNAL_LOCATIONS_HIGH = [ 28 + (i * 27 ) for i in range( 0, 6)]
SYNC_SIGNAL_LOCATIONS_LOW  = [ 14 + (i * 27 ) for i in range( 0, 6)]

## When searching for preamble - search in this range
SCAN_RANGE = range( -5, 18)
lastPreambleOffset = 0 ## Cache result

## Bit value of 1 above this 'luma' level, 0 below
THRESHOLD = 80

## Originally Closed Captions supported 5 channels, CC1, CC2, CC3, CC4
## But since nobody ever used CC3 and CC4, these were dropped and extra
## features were added, like the preamble stuff and the 'special' multi-byte
## characters.
##
## These late additions to the Closed caption standard are what make it slightly
## more cumbersome to decode than you would otherwise expect.

CCDict = {
    0x00: '', # Special - included here to clear a few things up
    0x20: ' ', 0x21: '!', 0x22: '"', 0x23: '#',
    0x24: '$', 0x25: '%', 0x26: '&', 0x27: "'", 
    0x28: "(", 0x29: ")", 0x2A: 'á', 0x2B: '+', 
    0x2C: ',', 0x2D: '-', 0x2E: '.', 0x2F: '/',
    0x3A: ':', 0x3B: ';', 0x3C: '<', 0x3D: '=', 
    0x3E: '>', 0x3F: '?', 0x40: '@', 0x5B: '[', 
    0x5C: 'é', 0x5D: ']', 0x5E: 'í', 0x5F: 'ó', 
    0x60: 'ú', 0x7B: 'ç', 0x7C: '÷', 0x7D: 'Ñ', 
    0x7E: 'ñ', 0x7F: '|',  # TODO 0x7F is Solid block    
    }

## Populate standard ASCII codes
for nrange in [ (0x41, 0x5B), (0x61, 0x7B), (0x30, 0x3A) ]: # ASCII ranges that are shared
    for i in xrange(nrange[0], nrange[1]):
        CCDict[i] = chr(i)

## Two byte chars
specialChars = {
    0x30: '®', 0x31: '°', 0x32: '½', 0x33: '¿', 
    0x34: 'TM', #TODO - resolve unicode symbol TM
    0x35: 'c',  #TODO - resolve unicode symbol cents
    0x36: '£',  0x37: '♪',
    0x38: 'a',  #TODO - with Grave accent
    0x39: ' ',  #TODO - Transparent space
    0x3A: 'e',  #TODO - with Grave accent
    0x3B: 'a',  0x3C: 'e', 0x3D: 'i', 0x3E: 'o', 0x3F: 'u',  #TODO - All with circumflex
}

CC1SpecialChars = dict(((0x11, a), b) for (a, b) in specialChars.items() )
CC2SpecialChars = dict(((0x19, a), b) for (a, b) in specialChars.items() )
AllSpecialChars = dict( CC1SpecialChars.items() + CC2SpecialChars.items() )

ControlCodes = {
    (0x14, 0x20) : 'Resume Caption Loading', 
    (0x14, 0x21) : 'Backspace', 
    (0x14, 0x22) : 'Reserved (Alarm Off)', 
    (0x14, 0x23) : 'Reserved (Alarm On)', 
    (0x14, 0x24) : 'Delete to End Of Row', 
    (0x14, 0x25) : 'Roll-Up Captions-2 Rows', 
    (0x14, 0x26) : 'Roll-Up Captions-3 Rows', 
    (0x14, 0x27) : 'Roll-Up Captions-4 Rows',  
    (0x14, 0x28) : 'Flash On',
    (0x14, 0x29) : 'Resume Direct Captioning', 
    (0x14, 0x2A) : 'Text Restart', 
    (0x14, 0x2B) : 'Resume Text Display', 
    (0x14, 0x2C) : 'Erase Displayed Memory', 
    (0x14, 0x2D) : 'Carriage Return', 
    (0x14, 0x2E) : 'Erase Non-Displayed Memory', 
    (0x14, 0x2F) : 'End of Caption (flip memory)', 
    (0x17, 0x21) : 'Tab Offset 1', 
    (0x17, 0x22) : 'Tab Offset 2', 
    (0x17, 0x23) : 'Tab Offset 3', 
}

CC1ControlCodes = dict((( a[0],  a[1])                       , 'CC1 ' + b) for (a, b) in ControlCodes.items() )
CC2ControlCodes = dict((( a[0]==0x14 and 0x1C or 0x1F,  a[1]), 'CC2 ' + b) for (a, b) in ControlCodes.items() )

MidRowCodes = {
    (0x11, 0x20) : 'Mid-row: White',     (0x11, 0x21) : 'Mid-row: White Underline', 
    (0x11, 0x22) : 'Mid-row: Green',     (0x11, 0x23) : 'Mid-row: Green Underline', 
    (0x11, 0x24) : 'Mid-row: Blue',      (0x11, 0x25) : 'Mid-row: Blue Underline', 
    (0x11, 0x26) : 'Mid-row: Cyan',      (0x11, 0x27) : 'Mid-row: Cyan Underline', 
    (0x11, 0x28) : 'Mid-row: Red',       (0x11, 0x29) : 'Mid-row: Red Underline', 
    (0x11, 0x2A) : 'Mid-row: Yellow',    (0x11, 0x2B) : 'Mid-row: Yellow Underline', 
    (0x11, 0x2C) : 'Mid-row: Magenta',   (0x11, 0x2D) : 'Mid-row: Magenta Underline', 
    (0x11, 0x2E) : 'Mid-row: Italics',   (0x11, 0x2F) : 'Mid-row: Italics Underline',     
}

CC1MidRowCodes = dict((( 0x11,  a[1]), 'CC1 ' + b) for (a, b) in MidRowCodes.items() )
CC2MidRowCodes = dict((( 0x19,  a[1]), 'CC2 ' + b) for (a, b) in MidRowCodes.items() )

## Preamble for odd columns except where it isn't
preambleOdd = {
    0x40: 'Pre: White',         0x41: 'Pre: White Underline', 
    0x42: 'Pre: Green',         0x43: 'Pre: Green Underline', 
    0x44: 'Pre: Blue',          0x45: 'Pre: Blue Underline', 
    0x46: 'Pre: Cyan',          0x47: 'Pre: Cyan Underline',
    0x48: 'Pre: Red',           0x49: 'Pre: Red Underline', 
    0x4A: 'Pre: Yellow',        0x4B: 'Pre: Yellow Underline',
    0x4C: 'Pre: Magenta',       0x4D: 'Pre: Magenta Underline',
    0x4E: 'Pre: White Italics', 0x4F: 'Pre: White Italics Underline', 
    0x50: 'Pre: Indent 0',      0x51: 'Pre: Indent 0 Underline',
    0x52: 'Pre: Indent 4',      0x53: 'Pre: Indent 4 Underline',
    0x54: 'Pre: Indent 8',      0x55: 'Pre: Indent 8 Underline', 
    0x56: 'Pre: Indent 12',     0x57: 'Pre: Indent 12 Underline',
    0x58: 'Pre: Indent 16',     0x59: 'Pre: Indent 16 Underline', 
    0x5A: 'Pre: Indent 20',     0x5B: 'Pre: Indent 20 Underline',     
    0x5C: 'Pre: Indent 24',     0x5D: 'Pre: Indent 24 Underline',     
    0x5E: 'Pre: Indent 28',     0x5F: 'Pre: Indent 28 Underline',     
}

evenPreamble = dict((a + 0x20,  b) for (a, b) in preambleOdd.items())

## Columns headings 
CC1preableCols = [0x11, 0x11, 0x12, 0x12, 0x15, 0x15, 0x16, 0x16, 0x17, 0x17, 0x10, 0x13, 0x13, 0x14, 0x14]
CC2preableCols = [0x19, 0x19, 0x1A, 0x1A, 0x1D, 0x1D, 0x1E, 0x1E, 0x1F, 0x1F, 0x18, 0x1B, 0x1B, 0x1C, 0x1C]

# Note candence change on column 11
colPreableTypes = [preambleOdd, evenPreamble, preambleOdd, evenPreamble, preambleOdd, evenPreamble, preambleOdd, evenPreamble, 
                   preambleOdd, evenPreamble, preambleOdd, preambleOdd, evenPreamble, preambleOdd, evenPreamble ] 

def CCPreableTable():
    '''Function generated due to complexity'''
    table = dict()
    for col, val in enumerate( colPreableTypes ):
        for (rowcode, text) in val.items():
            table[ (CC1preableCols[col], rowcode )] = 'CC1 ' + text + (' row %d' % (col+1))
            table[ (CC2preableCols[col], rowcode )] = 'CC2 ' + text + (' row %d' % (col+1))
    return table            

AllCCControlCodes = dict( CCPreableTable().items() + CC1ControlCodes.items() + CC2ControlCodes.items() +
                          CC1MidRowCodes.items() + CC2MidRowCodes.items() )

def isEmpty( byte1, byte2 ):
    return byte1 == 0 and byte2 == 0

def isControlCode( byte1,  byte2 ):
    return AllCCControlCodes.get( ( byte1, byte2 ) )
    
def memoize(f):
    ''' Memoization for performance on inner loop'''
    class memodict(dict):
        def __getitem__(self, *key):
            return dict.__getitem__(self, key)

        def __missing__(self, key):
            ret = self[key] = f(*key)
            return ret

    return memodict().__getitem__

@memoize
def decodeBytePair( byte1,  byte2 ):
    ''' Decode a pair of bytes'''
    controlcode = ( byte1, byte2 )
    if AllCCControlCodes.get( controlcode ):
        return AllCCControlCodes.get( controlcode )
    if AllSpecialChars.get( controlcode ):
        return AllSpecialChars.get( controlcode )        
    return '' + byteToChar( byte1 ) + byteToChar( byte2 )

def byteToChar( byte ):        
    ''' Take an int, and map it to a basic CC character or return ????(Hexcode) on failure'''
    return CCDict.get(byte , '????(%02x)' % byte )

def getPixel( image, x, y ):
    r, g, b, _ = cv.Get2D( image, y, x )
    return ( r + g + b ) / 3

@memoize
def bitsToByte( b6, b5, b4, b3, b2, b1, b0 ):
    return int( '%d%d%d%d%d%d%d' % ( b6, b5, b4, b3, b2, b1, b0 ),  2 )

def decodeByte( image, bitLocations,  samplesize, rownum,  offset=0 ):
    ''' Decode a single byte from a closed caption images
         row - is an array of luma values
         bitLocations - where to start sampling for each bit
         samplesize - how many pixels to average for each bit '''
    byte = []
    for pixel in bitLocations:
        val = sum( getPixel( image, i + offset, rownum ) for i in range(pixel, pixel+samplesize) ) / samplesize
        byte.append( val > THRESHOLD )
    # TODO - parity checking here
    return bitsToByte( byte[6], byte[5], byte[4], byte[3], byte[2], byte[1],  byte[0] )

def decodeRow( image , samplesize=3, rownum = 1, offset=0 ):
    ''' Attempt to pull two bytes worth of CC values out of a passed row of luma values
          samplesize - how many pixels wide to read each bit (Noise/drop-out reduction) 
          rownum     - which row of video to read as line 21 (typically row 1) '''
    return (decodeByte( image, BYTE1_LOCATIONS, samplesize, rownum, offset ), decodeByte( image, BYTE2_LOCATIONS, samplesize, rownum, offset ))

def isCCPresent( image, rownum=1 ):
    ''' Looks for the sine CC timing signal at the start of a row '''
    def scanPreamble( image, rownum,  offset ):
        for loc in SYNC_SIGNAL_LOCATIONS_HIGH:
            if getPixel(image, loc + offset, rownum ) < THRESHOLD:
                return False
        for loc in SYNC_SIGNAL_LOCATIONS_LOW:
            if getPixel(image, loc + offset, rownum ) > THRESHOLD:
                return False
        return True
    
    global lastPreambleOffset
    if scanPreamble( image,  rownum, lastPreambleOffset ): 
        return lastPreambleOffset #Optimisation - assume preamble doesn't drift
    for offset in SCAN_RANGE:
        if scanPreamble( image,  rownum, offset ):
            lastPreambleOffset = offset          
            ## Found a match - but let's optimize
            for optim in range(12):
                if not scanPreamble( image,  rownum, offset + optim ):
                    ## Found a negative match, let's take the middle of the range
                    lastPreambleOffset = int( offset + ( 0.5 * optim  ) )
                    break
            return lastPreambleOffset
    return None
    
def decodeCaptionsRaw( imageList, mergeText=False, deleteImageAfterProcessing=True ):
    '''Debug output, show the frame caption codes and frame numbers'''
    buff = '' ## CC Buffer
    frame = 0
    for imageName in imageList:
        img = cv.LoadImage( imageName )
        offset = isCCPresent( img )
        if offset is None:
            print '%i skip - no preamble' % frame            
        else: 
            byte1, byte2 = decodeRow( img, offset=offset )
            code = decodeBytePair( byte1, byte2 )
            empty = isEmpty( byte1, byte2 )                
            if not isControlCode( byte1, byte2 ) and not empty:
                if mergeText:
                    buff += code
                else:
                    print '%i Text:%s' % (frame, code )                
            elif buff:
                print '%i Text:%s' % (frame, buff )
                buff = ''
            if isControlCode( byte1, byte2 ):
                print '%i %s' % (frame, code )
        frame+=1
        if deleteImageAfterProcessing:
            os.unlink( imageName )    
    
def decodeImageListToSRT( imagelist, framesPerSecond, deleteImageAfterProcessing=True ):
    ''' Decode a passed list of images to a stream of SRT subtitles. Assumes Pop-on format closed captions 
         imagelist                  - list of image file paths
         framesPerSecond            - how many fps is the passed list of images
         deleteImageAfterProcessing - delete the image file after we have done processing it'''
         
    def SRTtimeStamp( frames, fps ):
        '''Returns an SRT format timestamp
            frames - frame offset from start of the file
            fps - frames per second in this file'''
        seconds = frames/fps
        milliseconds = int ( ( seconds - int(seconds) ) * 1000 )
        hours = int( seconds / 3600 )
        minutes = int( ( seconds - ( 3600 * hours ) ) / 60 )
        secondsDisp = seconds - ( ( minutes * 60 ) + ( hours * 3600 ) )
        return '%02d:%02d:%02d,%03d' % ( hours, minutes,  secondsDisp,  milliseconds )

    def dumpSrtCaption( text, start, end, fps ):
        '''Display an SRT format closed caption
            text  - the text to display   start - the starting frame
            end   - the ending frame      fps   - the fps of the file'''
        print SRTtimeStamp( start, fps ) + ' --> ' + SRTtimeStamp( end, fps )
        print text
        print '' 
    
    offscreen = '' ## CC Buffer
    onscreen  = ''  ## CC Buffer
    prevcode  = None 
    frame     = 0
    subTitleStartFrame = 0
    for imageName in imagelist:
        img = cv.LoadImage( imageName )
        # DECODE
        offset = isCCPresent( img )
        if offset is None:
            byte1 = 0 ## Skip where no sine preamble detected
            byte2 = 0
            code = None
        else:
            byte1, byte2 = decodeRow( img, offset=offset )
            code = decodeBytePair( byte1, byte2 )
            control = isControlCode( byte1, byte2 )
        # PROCESS
        if isEmpty( byte1, byte2 ):
            pass
        elif not control:            
            offscreen += code  ## Must be text
        elif control and code != prevcode :
            if 'End of Caption' in code:
                onscreen = offscreen
                offscreen = ''
                subTitleStartFrame = frame
            elif onscreen and 'Erase Displayed Memory' in code:
                dumpSrtCaption( onscreen, subTitleStartFrame, frame, framesPerSecond ) 
                onscreen = ''
            elif offscreen and offscreen[-1:] != '\n':
                offscreen +='\n' ## Some random command code. Assume it's just a newline
        # CLEANUP
        prevcode = code
        frame+=1
        if deleteImageAfterProcessing:
            os.unlink( imageName )       

def streamDecodeFileList( inputFile, tempdir ):
    ''' Returns a generator of image file names based on ffmpeg decoding the top 10 lines of the passed inputFile '''
    ## This is a bit hacky. Basically run ffmpeg in a subprocess, and keep processing the produced tiffs until 
    ## ffmpeg closes and we run out of frames.
    pathToTempDir = tempfile.mkdtemp(dir=tempdir)
    tempfileNameStyle = 'ccdecode%07d.tif'
    ffmpegCmd = PATH_TO_FFMPEG + ( ' -i "%s"' % inputFile ) + ' -vf "crop=iw:10:0:0"' + ' -f image2 "' + pathToTempDir + os.sep + tempfileNameStyle + '"'
    print ffmpegCmd
    with open( os.devnull, 'wb') as devnull:
        fpid = subprocess.Popen(ffmpegCmd, stderr=devnull)
        fileNum = 1
        while fpid.poll() is None:
            nextFileName = pathToTempDir + os.sep + ( tempfileNameStyle % fileNum )
            if os.path.exists( nextFileName ): 
                fileNum+=1
                nextNextFileName = pathToTempDir + os.sep + ( tempfileNameStyle % fileNum )
                if os.path.exists( nextNextFileName ):                
                    yield nextFileName        
                else:
                    time.sleep( 1 ) ## We are at the head, so wait a moment for the file to complete writing
                    yield nextFileName
            else:
                time.sleep( 1 ) ## Sleep to let ffmpeg get ahead of us
        ## FFMpeg must have exited - race to the conclusion
        nextFileName = pathToTempDir + os.sep + ( tempfileNameStyle % fileNum )
        while os.path.exists( nextFileName ):
            yield nextFileName
            fileNum+=1
            nextFileName = pathToTempDir + os.sep + ( tempfileNameStyle % fileNum )    
    os.rmdir( pathToTempDir )
            
def main():
    fileListGenerator = streamDecodeFileList( 'C:\\Video\\InputVideo.mov', 'C:\\Temp' )
    ## decodeCaptionsRaw( fileListGenerator ) ## For debugging
    decodeImageListToSRT( fileListGenerator, 29.97, deleteImageAfterProcessing=True )

## Adjust content of the main method before running
main()

